#ifndef FLYING_CAM_H
#define FLYING_CAM_H

//#include "common.h"

//#include <windows.h>

#include "util.h"
#include "mathlib/vector.h"
//#include <gl/GL.h>



class FlyingCam
{
	vector3 m_pos, m_look, m_up, m_right;

	float m_xRot, m_yRot;

	float m_frontSpeed, m_sideSpeed;

public:
	enum KEY_INPUT
	{
		UP, DOWN, LEFT, RIGHT, NO_KEY
	};

	const vector3 &GetPos() const {return m_pos;}

	void init(const vector3 &pos, float frontspeed, float sidespeed)
	{
		// cam setting
		m_xRot = 180;
		m_yRot = 0;

		// init
		m_pos = pos;
		look(-900, 0);


		m_frontSpeed = frontspeed;
		m_sideSpeed = sidespeed;
	}

	vector3 preMove(KEY_INPUT key)
	{
		vector3 p = m_pos;

		switch(key)
		{
		case UP: // up
			p += m_look * m_frontSpeed;
			break;

		case DOWN: // down
			p -= m_look * m_frontSpeed;
			break;

		case LEFT: // left
			p -= m_right * m_sideSpeed;
			break;

		case RIGHT: // right
			p += m_right * m_sideSpeed;
			break;
		}

		return p;
	}

	void move(KEY_INPUT key) // ġ
	{
		switch(key)
		{
		case UP: // up
			m_pos += m_look * m_frontSpeed;
			break;

		case DOWN: // down
			m_pos -= m_look * m_frontSpeed;
			break;

		case LEFT: // left
			m_pos -= m_right * m_sideSpeed;
			break;

		case RIGHT: // right
			m_pos += m_right * m_sideSpeed;
			break;
		}
	}
	
	void move(float xoffset, int zoffset)
	{
		m_pos.x += xoffset;
		m_pos.z += zoffset;
	}

	void look(int offsetx, int offsety) // ġ 
	{
		m_yRot -= offsetx / 10.0f;
		m_xRot -= offsety / 10.0f;

		if(m_yRot > 360)
			m_yRot = 0;
		if(m_yRot < 0)
			m_yRot = 360;

		if(m_xRot > 360)
			m_xRot = 0;
		if(m_xRot < 0)
			m_xRot = 360;

		float mark = (90 < m_xRot && m_xRot < 270) ? -1 : 1; // org

		m_look.x = sin(degtorad(m_yRot)) * cos(degtorad(m_xRot));
		m_look.y = -sin(degtorad(m_xRot));
		m_look.z = cos(degtorad(m_xRot)) * cos(degtorad(m_yRot));

		vector3 left = m_look * vector3(0, 1, 0);
		m_up = m_look * left;

		m_right = left * -1;

		m_up *= mark;
		m_right *= mark;

		m_look.norm();
		m_right.norm();
		m_up.norm();

//		DEBUG_LOG("pos: %f/%f/%f, look: %f/%f/%f up: %f/%f/%f", m_pos.x, m_pos.y, m_pos.z, m_look.x, m_look.y, m_look.z, m_up.x, m_up.y, m_up.z);
	}

	void lookAt()
	{
		gluLookAt(m_pos.x, m_pos.y, m_pos.z, 
			m_pos.x + m_look.x, m_pos.y + m_look.y, m_pos.z + m_look.z, 
			m_up.x, m_up.y, m_up.z);

	}

};

#endif