#include "stdafx.h"
#include "PlaneSim.h"


PlaneSim::PlaneSim()
{
	m_oldX = 0;
	m_oldY = 0;

	m_isBirdView = false;
}

void PlaneSim::OnCreate()
{
	// render context
	m_hdc = GetDC(m_hwnd);

	PIXELFORMATDESCRIPTOR pfd;

	memset(&pfd, 0, sizeof(pfd));

	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER | PFD_GENERIC_ACCELERATED;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = COLOR_BITS;
	pfd.cDepthBits = DEPTH_BITS;
	pfd.cStencilBits = STENCIL_BITS;

	SetPixelFormat(m_hdc, ChoosePixelFormat(m_hdc, &pfd), &pfd);

	HGLRC hrc = wglCreateContext(m_hdc);
	wglMakeCurrent(m_hdc, hrc);


	// init gl
	// 
	glEnable(GL_DEPTH_TEST);
	glFrontFace(GL_CCW);	
	glEnable(GL_CULL_FACE);	

	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );

	glClearColor(0, 0, 0, 0);

	glShadeModel(GL_SMOOTH);

	glLineWidth(2);
	glPointSize(10);

	// light
	// 
	GLfloat  whiteLight[] = { 0.5f, 0.5f, 0.5f, 1.0f };
	GLfloat  ambientLight[] = { 0.1f, 0.1f, 0.1f, 1.0f };
	GLfloat  diffuseLight[] = { 1.0f, 1.0f, 1.0f, 1.0f };
	m_lightPos.set(Terrain::TERRAIN_MAX_X * Terrain::TILE_SIZE / 2, 500, Terrain::TERRAIN_MAX_Z * Terrain::TILE_SIZE / 2);
	GLfloat	 lightPos[] = {m_lightPos.x, m_lightPos.y, m_lightPos.z, 1 };

	glEnable(GL_LIGHTING);
	//	glLightModelfv(GL_LIGHT_MODEL_AMBIENT,whiteLight);
	glLightfv(GL_LIGHT0,GL_AMBIENT,ambientLight);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuseLight);
	glLightfv(GL_LIGHT0,GL_POSITION,lightPos);
	glEnable(GL_LIGHT0);

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);


#ifdef _DEBUG
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
#else
	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
#endif

	m_cam.init(vector3(100, 100, 100), MOVE_SPEED, MOVE_SPEED);
	
	m_terrain.Init();


	// load models
	Bitmap toon;

	if(toon.load("textures/toon.bmp"))
	{
		m_toonTexture.create(toon);
		
	//	m_models[0].Load("models/sphere5.dat", &m_toonTexture, 150, 150, 10, 50, 0.01);
		m_models[0].Load("models/tie.dat", &m_toonTexture, 150, 150, 10, 50, 0.005);
		m_models[1].Load("models/x_wing.dat", &m_toonTexture, 400, 400, 10, 100, -0.007);
		m_models[2].Load("models/space_station.dat", &m_toonTexture, 600, 600, 10, 100, 0.005);
	}
	else
	{
		WARN_LOG("cannot load terrain texture!");
	}

}

void PlaneSim::OnIdle()
{
	Sleep(1);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	if(m_isBirdView)
	{
		vector3 pos(Terrain::TERRAIN_MAX_X * Terrain::TILE_SIZE / 2, 1000, Terrain::TERRAIN_MAX_Z * Terrain::TILE_SIZE * 1.3);
		vector3 look(0, -0.5, -0.5);
		look.norm();
		vector3 eye = pos + look;

		gluLookAt(pos.x, pos.y, pos.z, eye.x, eye.y, eye.z, 0, 1, 0);

		// rendering me
		const vector3 &me = m_cam.GetPos();

		glColor3f(1, 0, 0);
		glPushMatrix();
		glTranslatef(me.x, me.y, me.z);
		glutSolidSphere(2.0f, 30, 30);
		glPopMatrix();
	}
	else
	{
		m_cam.lookAt();
	}

	glDisable(GL_LIGHTING);

	// draw axis
	DrawAxis(1000);

	// light
	glPushMatrix();
	glTranslatef(m_lightPos.x, m_lightPos.y, m_lightPos.z);
	glColor3f(1, 1, 1);
	glutSolidSphere(10.0f, 30, 30);
	glPopMatrix();

	// render terrain
	const vector3 &mypos = m_cam.GetPos();
	m_terrain.Render(mypos);

	// test render model
	glColor3f(1, 1, 1);

	//glColor3f(0.2, 0.2, 0.25);
	m_models[0].Render(m_cam, m_terrain, m_lightPos);

	//glColor3f(0.7, 0.5, 0.4);
	m_models[1].Render(m_cam, m_terrain, m_lightPos);

	//glColor3f(0.7, 0.5, 0.8);
	m_models[2].Render(m_cam, m_terrain, m_lightPos);

	glDisable(GL_TEXTURE_2D);

	// test sphere
	for(int i = 0; i < 10; i++)
	{
		glColor3f(0.1 * i, 0.2 * (i / 2), 0.3 * i / 4);

		float x = 100 * i;
		float z = 50 * i;
		float y = m_terrain.GetHeight(x, z);

		glPushMatrix();
		glTranslatef(x, y, z);
		glutSolidSphere(3.0f, 30, 30);
		glPopMatrix();
	}
	
	SwapBuffers(m_hdc);
}

void PlaneSim::DrawAxis(int len)
{
	glBegin(GL_LINES);
	// x, red
	glColor3f(1, 0, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(len, 0, 0);

	// y, green
	glColor3f(0, 1, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(0, len, 0);

	// z, blue
	glColor3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 0, len);

	glEnd();

}

LRESULT PlaneSim::WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int wmId, wmEvent;

	switch (message)
	{
	case WM_CREATE:
		{
			m_hwnd = hWnd;
			OnCreate();
		}
		break;

	case WM_SIZE:
		{
			int width = LOWORD(lParam);
			int height = HIWORD(lParam);

			if(height == 0)
				height = 1;

			m_width = width;
			m_height = height;

			glViewport(0, 0, width, height);

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();

			gluPerspective(45, (float)width / height, 0.1, 10000);

			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

		}
		break;

	case WM_KEYDOWN:
		{
			FlyingCam::KEY_INPUT key = FlyingCam::NO_KEY;

			static const int VK_1 = 0x31;
			static const int VK_W = 0x57, VK_S = 0x53, VK_A = 0x41, VK_D = 0x44;

			// for birdview move
			int xoffset = 0;
			int zoffset = 0;

			switch(wParam)
			{
			case VK_1:
				m_isBirdView = !m_isBirdView;
				break;

			case VK_W: // up
				key = FlyingCam::UP;
				zoffset = -MOVE_SPEED;
				break;

			case VK_S: // down
				key = FlyingCam::DOWN;
				zoffset = MOVE_SPEED;
				break;

			case VK_A: // left
				key = FlyingCam::LEFT;
				xoffset = -MOVE_SPEED;
				break;

			case VK_D: // right
				key = FlyingCam::RIGHT;
				xoffset = MOVE_SPEED;
				break;
			}

			if(key != FlyingCam::NO_KEY)
			{				 
				// 浹üũ
				vector3 prep;

				if(m_isBirdView)
				{
					prep = m_cam.GetPos();
					prep.x += xoffset;
					prep.z += zoffset;
				}
				else
				{
					prep = m_cam.preMove(key);
				}

				float minheight = m_terrain.GetHeight(prep.x, prep.z) + 1;

				if((prep.x < 0 || prep.z < 0) || ((prep.x > 0 && prep.z > 0) && minheight < prep.y))
				{
					sphere s(prep, 0.01f);
					bool intersect = false;

					for(int i = 0; i < N_MODELS; i++)
					{
						if(m_models[i].GetSphere().intersects(s))
						{
							intersect = true;
							break;
						}
					}

					if(!intersect)
					{
						if(m_isBirdView)
						{
							m_cam.move(xoffset, zoffset);
						}
						else
						{
							m_cam.move(key);
						}
					}
				}
			}
		}	
		break;

	case WM_MOUSEMOVE:
		{
			if(m_isBirdView)
				break;

			static bool ignore = false;

			if(ignore)
			{
				ignore = false;
				break;
			}

			int xPos = GET_X_LPARAM(lParam);
			int yPos = GET_Y_LPARAM(lParam); 

			m_cam.look(xPos - m_oldX, yPos - m_oldY);

			m_oldX = xPos;
			m_oldY = yPos;

			ignore = true;
			POINT pt = {m_width / 2, m_height / 2};
			ClientToScreen(hWnd, &pt);
			SetCursorPos(pt.x, pt.y);
			m_oldX = m_width / 2;
			m_oldY = m_height / 2;
		}
		break;

	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

