#include "stdafx.h"
#include <gl/GL.h>
#include "include/glut.h"

#include "model.h"


void Model::Render(const FlyingCam &cam, const Terrain &terrain, vector3 lightpos)
{
	// rotated pos
	float x = m_rotRadius * cos(m_rotAngle);
	float z = m_rotRadius * sin(m_rotAngle);
	vector3 pos(m_basePos.x + x, 0, m_basePos.z + z);
	pos.y = terrain.GetHeight(pos.x, pos.z) + m_yOffset - m_lowerY;

	// 浹 Ʈ .
 	sphere cams(cam.GetPos(), 0.01);
	sphere nexts(pos, m_sphere.r);
	if(!cams.intersects(nexts))
	{
		m_sphere.p = pos;

		m_rotAngle += m_rotSpeed;

		if(m_rotAngle > 360)
			m_rotAngle = 0;
		else if(m_rotAngle < 0)
			m_rotAngle = 360;
	}

	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);

	//glEnable(GL_LIGHTING);
	//glDisable(GL_TEXTURE_2D);

	m_pToonTexture->bind();

	glPushMatrix();
	glTranslatef(m_sphere.p.x, m_sphere.p.y, m_sphere.p.z);
	//glRotatef(270, 1, 0, 0);

	matrix44 m;
	m.translate(m_sphere.p);
	
	static const float BLACK_TEX_S_COORD = 0;

	glBegin(GL_TRIANGLES);
	for(int i = 0; i < m_faces.size(); i++)
	{
		const vector3 &idx = m_faces[i];
		const vector3 &a = m_verts[idx.x];
		const vector3 &b = m_verts[idx.y];
		const vector3 &c = m_verts[idx.z];

		const vector3 &na = m_norms[idx.x];
		const vector3 &nb = m_norms[idx.y];
		const vector3 &nc = m_norms[idx.z];

		// calc intensity
		vector3 reala = (m * a);
		vector3 realb = (m * b);
		vector3 realc = (m * c);

		vector3 lighta = lightpos - reala;
		vector3 lightb = lightpos - realb;
		vector3 lightc = lightpos - realc;

		lighta.norm();
		lightb.norm();
		lightc.norm();

		float intensitya = na.dot(lighta);
		float intensityb = nb.dot(lightb);
		float intensityc = nc.dot(lightc);


		if(intensitya < 0)
			intensitya = BLACK_TEX_S_COORD;
		if(intensityb < 0)
			intensityb = BLACK_TEX_S_COORD;
		if(intensityc < 0)
			intensityc = BLACK_TEX_S_COORD;

		//DEBUG_LOG("intensity: %f, %f, %f", intensitya, intensityb, intensityc);

		glTexCoord2f(intensitya, 0);
		glNormal3f(na.x, na.y, na.z);
		glVertex3f(a.x, a.y, a.z);

		glTexCoord2f(intensityb, 0);
		glNormal3f(nb.x, nb.y, nb.z);
		glVertex3f(b.x, b.y, b.z);

		glTexCoord2f(intensityc, 0);
		glNormal3f(nc.x, nc.y, nc.z);
		glVertex3f(c.x, c.y, c.z);
		
	}
	glEnd();

	//// draw norms
	//glColor3f(1, 0, 0);

	//glDisable(GL_LIGHTING);
	//glBegin(GL_LINES);
	//for(int i = 0; i < m_faces.size(); i++)
	//{
	//	const vector3 &idx = m_faces[i];
	//	const vector3 &a = m_verts[idx.x];
	//	const vector3 &b = m_verts[idx.y];
	//	const vector3 &c = m_verts[idx.z];

	//	const vector3 na = a + m_norms[idx.x] * 10;
	//	const vector3 nb = b + m_norms[idx.y] * 10;
	//	const vector3 nc = c + m_norms[idx.z] * 10;

	//	glVertex3f(na.x, na.y, na.z);
	//	glVertex3f(a.x, a.y, a.z);

	//	glVertex3f(nb.x, nb.y, nb.z);
	//	glVertex3f(b.x, b.y, b.z);

	//	glVertex3f(nc.x, nc.y, nc.z);
	//	glVertex3f(c.x, c.y, c.z);
	//}
	//glEnd();
	//glEnable(GL_LIGHTING);

	glPopMatrix();

#ifdef _DEBUG
	glPushMatrix();
	glTranslatef(m_sphere.p.x, m_sphere.p.y, m_sphere.p.z);
	glEnable(GL_BLEND);
	glColor4f(1, 1, 1, 0.5f);
	glutSolidSphere(m_sphere.r, 30, 30);
	glDisable(GL_BLEND);
	glPopMatrix();
#endif
	
}

bool Model::Load(const string &filename, Texture *toonTexture, float x, float z, float yoffset, float radius, float speed)
{
	m_pToonTexture = toonTexture;

	m_basePos.x = x;
	m_basePos.z = z;
	m_yOffset = yoffset;

	m_rotRadius = radius;
	m_rotSpeed = speed;
	m_rotAngle = 0;


	FILE *fp = fopen(filename.c_str(), "rt");

	printf("load model file: %s\n", filename.c_str());

	if(fp == 0)
	{
		printf("cannot open file: %s\n", filename.c_str());
		return false;
	}

	char seps[] = " \t";
	const int BUF_LEN = 1024;
	char buf[BUF_LEN] = {0};
	char buf2[BUF_LEN] = {0};
	char *ptok;
	MODEL_PARSE_STATE state = N_VERTS;

	int nVerts = 0;
	int nFaces = 0;

	matrix44 rotm;
	rotm.rotate_x(degtorad(270));

	while(fgets(buf, BUF_LEN, fp) != 0)
	{
		if(strcmp(buf, "\n") == 0)
		{
		}
		else if(state == N_VERTS && strncmp(buf, "VERTEX", 6) == 0)
		{
			sscanf(buf, "%s = %d", buf2, &nVerts);
			state = VERTS;
		}
		else if(strncmp(buf, "FACE", 4) == 0)
		{
			state = FACES;
		}
		else if(state == VERTS)
		{
			double x, y, z;

			ptok = strtok(buf, seps);
			x = atof(ptok);
			ptok = strtok(0, seps);
			y = atof(ptok);
			ptok = strtok(0, seps);
			z = atof(ptok);

			// transform
			vector3 v(x, y, z);
			vector3 v2 = rotm * v;

			m_verts.push_back(v2);
		}
		else if(state == FACES)
		{
			int x, y, z;

			ptok = strtok(buf, seps);
			x = atoi(ptok);
			ptok = strtok(0, seps);
			y = atoi(ptok);
			ptok = strtok(0, seps);
			z = atoi(ptok);

			m_faces.push_back(vector3(x, y, z));
		}
	}

	// calc norms
	vector<vector3> facenorms;

	for(int i = 0; i < m_faces.size(); i++)
	{
		const vector3 &idx = m_faces[i];
		const vector3 &a = m_verts[idx.x];
		const vector3 &b = m_verts[idx.y];
		const vector3 &c = m_verts[idx.z];

		vector3 ab = b - a;
		vector3 ac = c - a;

		vector3 n = ab * ac;
		n.norm();

		facenorms.push_back(n);
	}

	for(int i = 0; i < m_verts.size(); i++)
	{
		vector3 n(0, 0, 0);

		for(int j = 0; j < m_faces.size(); j++)
		{
			// if face j is adjacent to vert i
			const vector3 &face = m_faces[j];
			if(face.x == i || face.y == i || face.z == i)
			{
				n += facenorms[j];
			}
		}

		n.norm();
		m_norms.push_back(n);
	}


	// get lower point & bound sphere
	float maxlensq = -1;
	float lower_y = 99999;

	for(int i = 0; i < m_verts.size(); i++)
	{
		float lensq = m_verts[i].lensquared();
		
		if(m_verts[i].y < lower_y)
			lower_y = m_verts[i].y;

		if(lensq > maxlensq)
			maxlensq = lensq;
	}

	m_lowerY = lower_y;
	m_sphere.r = sqrt(maxlensq);

	INFO_LOG("model loaded. verts: %d, faces: %d, lowery: %f, radius: %f\n", m_verts.size(), m_faces.size(), m_lowerY, m_sphere.r);

	return true;
}