#include "stdafx.h"

#include "terrain.h"


void Terrain::Init()
{
	Bitmap sand;

	if(sand.load("textures/sand.bmp"))
	{
		m_terrainTexture.create(sand);
	}
	else
	{
		WARN_LOG("cannot load terrain texture!");
	}

	// create terrain
	for(int x = 0; x < TERRAIN_MAX_X; x++)
	{
		for(int z = 0; z < TERRAIN_MAX_Z; z++)
		{
			m_heights[x][z] = 0;
		}
	}

	float c1, c2, c3, c4;

	c1 = (float)random();
	c2 = (float)random();
	c3 = (float)random();
	c4 = (float)random();

	DivideGrid( 0, 0, TERRAIN_MAX_X - 1 , TERRAIN_MAX_Z - 1, c1, c2, c3, c4);

	// calc face norms
	for(int x = 0; x < TERRAIN_MAX_X; x++)
	{
		for(int z = 0; z < TERRAIN_MAX_Z; z++)
		{
			vector3 v00(x * TILE_SIZE, m_heights[x][z], z * TILE_SIZE);
			vector3 v01(x * TILE_SIZE, m_heights[x][z+1], (z+1) * TILE_SIZE);
			vector3 v11((x+1) * TILE_SIZE, m_heights[x+1][z+1], (z+1) * TILE_SIZE);
			vector3 v10((x+1) * TILE_SIZE, m_heights[x+1][z], z * TILE_SIZE);

			vector3 n = (v01 - v00) * (v10 - v00);
			n.norm();

			m_norms[x][z] = n;
		}
	}
}

float Terrain::Displace(float num)
{
	return ((float)random() - 0.5f);
}

void Terrain::DivideGrid(int x, int y, int width, int height, float c1, float c2, float c3, float c4)
{
	float Edge1, Edge2, Edge3, Edge4, Middle;
	int newWidth = width / 2;
	int newHeight = height / 2;

	if (width > 1 || height > 1)
	{	
		Middle = (c1 + c2 + c3 + c4) / 4 + Displace(newWidth + newHeight);	//Randomly displace the midpoint!
		Edge1 = (c1 + c2) / 2;	//Calculate the edges by averaging the two corners of each edge.
		Edge2 = (c2 + c3) / 2;
		Edge3 = (c3 + c4) / 2;
		Edge4 = (c4 + c1) / 2;

		if (Middle < 0)
		{
			Middle = 0;
		}
		else if (Middle > 1.0f)
		{
			Middle = 1.0f;
		}

		DivideGrid( x, y, newWidth, newHeight, c1, Edge1, Middle, Edge4);
		DivideGrid( x + newWidth, y, newWidth, newHeight, Edge1, c2, Edge2, Middle);
		DivideGrid( x + newWidth, y + newHeight, newWidth, newHeight, Middle, Edge2, c3, Edge3);
		DivideGrid( x, y + newHeight, newWidth, newHeight, Edge4, Middle, Edge3, c4);
	}
	else	
	{
		float c = (c1 + c2 + c3 + c4) / 4;

		m_heights[x][y] = c * MAX_HEIGHT;
	}
}


inline float Terrain::GetTileHeight(int x, int z)
{
	if(m_lodHeights[x][z] != -1)
		return m_lodHeights[x][z];
	
	return m_heights[x][z];
}


void Terrain::DrawTile(int xmin, int xmax, int zmin, int zmax, int lev)
{
	//DEBUG_LOG("lod %d %d, %d %d", xmin, xmax, zmin, zmax);

	if(xmax - xmin == 0 || zmax - zmin == 0)
		return;

	if(lev == 1)
	{
		// 簢  п ؼ height 
		int xdelta = xmax - xmin;
		for(int i = 0; i <= xdelta; i++)
		{	
			m_lodHeights[xmin + i][zmin] = (GetTileHeight(xmax, zmin) - GetTileHeight(xmin, zmin)) * i / xdelta + GetTileHeight(xmin, zmin);
			m_lodHeights[xmin + i][zmax] = (GetTileHeight(xmax, zmax) - GetTileHeight(xmin, zmax)) * i / xdelta + GetTileHeight(xmin, zmax);
		}

		int zdelta = zmax - zmin;
		for(int i = 0; i <= zdelta; i++)
		{
			m_lodHeights[xmin][zmin + i] = (GetTileHeight(xmin, zmax) - GetTileHeight(xmin, zmin)) * i / zdelta + GetTileHeight(xmin, zmin);
			m_lodHeights[xmax][zmin + i] = (GetTileHeight(xmax, zmax) - GetTileHeight(xmax, zmin)) * i / zdelta + GetTileHeight(xmax, zmin);
		}

		float y = 0;
		int repeat = xmax - xmin;

		//1
		glTexCoord2f(0, 0);
		DrawVert(xmin, zmin);

		glTexCoord2f(0, repeat);
		DrawVert(xmin, zmax);

		glTexCoord2f(repeat, repeat);
		DrawVert(xmax, zmax);

		//2
		glTexCoord2f(0, 0);
		DrawVert(xmin, zmin);

		glTexCoord2f(repeat, repeat);
		DrawVert(xmax, zmax);

		glTexCoord2f(repeat, 0);
		DrawVert(xmax, zmin);
	}
	else
	{
		int xmid = (xmax + xmin) / 2;
		int zmid = (zmax + zmin) / 2;

		DrawTile(xmin, xmid, zmin, zmid, lev - 1);
		DrawTile(xmin, xmid, zmid, zmax, lev - 1);
		DrawTile(xmid, xmax, zmin, zmid, lev - 1);
		DrawTile(xmid, xmax, zmid, zmax, lev - 1);
	}

}

inline void Terrain::DrawVert(int x, int z)
{
	float y = GetTileHeight(x, z);
//	float cr = (y / MAX_HEIGHT / 2) + (MAX_HEIGHT / 2);
	float cr = (y / MAX_HEIGHT);
	//glColor3f(cr, cr, cr);

	glColor3f(1, 1, 1);
	glNormal3f(m_norms[x][z].x, m_norms[x][z].y, m_norms[x][z].z);
	glVertex3f(x * TILE_SIZE, y, z * TILE_SIZE);

}

void Terrain::AddTileLevel(int xmin, int xmax, int zmin, int zmax, int lev)
{
	if(lev == 0)
		return;
	if(xmax - xmin <= 1 || zmax - zmin <= 1)
		return;

	int xmid = (xmax + xmin) / 2;
	int zmid = (zmax + zmin) / 2;

	m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmin, xmid, zmin, zmid)));
	m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmin, xmid, zmid, zmax)));
	m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmid, xmax, zmin, zmid)));
	m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmid, xmax, zmid, zmax)));

	AddTileLevel(xmin, xmid, zmin, zmid, lev - 1);
	AddTileLevel(xmin, xmid, zmid, zmax, lev - 1);
	AddTileLevel(xmid, xmax, zmin, zmid, lev - 1);
	AddTileLevel(xmid, xmax, zmid, zmax, lev - 1);
}


bool Terrain::CalcLODTerrain(int xmin, int xmax, int zmin, int zmax, int myposx, int myposz, int lev)
{
	if(xmax - xmin <= 1 || zmax - zmin <= 1)
		return false;

	int xmid = (xmax + xmin) / 2;
	int zmid = (zmax + zmin) / 2;

	int d = 4;

	if((xmin - d <= myposx && myposx <= xmax + d) && (zmin - d <= myposz && myposz <= zmax + d))
	{
		bool childdrew = false;

		childdrew |= CalcLODTerrain(xmin, xmid, zmin, zmid, myposx, myposz, lev + 1);
		childdrew |= CalcLODTerrain(xmin, xmid, zmid, zmax, myposx, myposz, lev + 1);
		childdrew |= CalcLODTerrain(xmid, xmax, zmin, zmid, myposx, myposz, lev + 1);
		childdrew |= CalcLODTerrain(xmid, xmax, zmid, zmax, myposx, myposz, lev + 1);

		if(!childdrew)
		{
			m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmin, xmid, zmin, zmid)));
			m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmin, xmid, zmid, zmax)));
			m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmid, xmax, zmin, zmid)));
			m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmid, xmax, zmid, zmax)));
		}

		return true;
	}
	else
	{
		m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmin, xmid, zmin, zmid)));
		m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmin, xmid, zmid, zmax)));
		m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmid, xmax, zmin, zmid)));
		m_drawSeqMap.insert(pair<int, DRAW_TILE>(lev, DRAW_TILE(xmid, xmax, zmid, zmax)));

		return false;
	}
}

void Terrain::Render(const vector3 &mypos)
{
	int myxpos = min(max(mypos.x / TILE_SIZE, 0), TERRAIN_MAX_X - 1);
	int myzpos = min(max(mypos.z / TILE_SIZE, 0), TERRAIN_MAX_Z - 1);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);

		
	/*
	// normal terrain
	float d = 0.1f;
	glBegin(GL_TRIANGLES);
	glColor3f(1, 1, 1);
	for(int x = 0; x < TERRAIN_MAX_X - 1; x++)
	{
		for(int z = 0; z < TERRAIN_MAX_Z - 1; z++)
		{
			glVertex3f(x * TILE_SIZE, m_heights[x][z] - d, z * TILE_SIZE);
			glVertex3f(x * TILE_SIZE, m_heights[x][z + 1] - d, (z + 1) * TILE_SIZE);
			glVertex3f((x + 1) * TILE_SIZE, m_heights[x + 1][z + 1] - d, (z + 1) * TILE_SIZE);

			glVertex3f(x * TILE_SIZE, m_heights[x][z] - d, z * TILE_SIZE);
			glVertex3f((x + 1) * TILE_SIZE, m_heights[x + 1][z + 1] - d, (z + 1) * TILE_SIZE);
			glVertex3f((x + 1) * TILE_SIZE, m_heights[x + 1][z] - d, z * TILE_SIZE);
		}
	}
	glEnd();
	*/

	
	// lod terrain
	for(int x = 0; x < TERRAIN_MAX_X; x++)
	{
		for(int z = 0; z < TERRAIN_MAX_Z; z++)
		{
			m_lodHeights[x][z] = -1;
		}
	}
	m_drawSeqMap.clear();

	m_terrainTexture.bind();

	glColor3f(1, 1, 1);
	glBegin(GL_TRIANGLES);
	CalcLODTerrain(0, TERRAIN_MAX_X - 1, 0, TERRAIN_MAX_Z - 1, myxpos, myzpos, 0);

	for(DrawSeqMap::iterator i = m_drawSeqMap.begin(); i != m_drawSeqMap.end(); i++)
	{
		DRAW_TILE &dt = i->second;

		if(dt.xmax - dt.xmin >= 2)
		{
			DrawTile(dt.xmin, dt.xmax, dt.zmin, dt.zmax, 3);
		}
		else
		{
			DrawTile(dt.xmin, dt.xmax, dt.zmin, dt.zmax, 1);
		}
	}

	glEnd();
	
	
	/*
	// draw norms
	glDisable(GL_LIGHTING);
	glColor3f(0, 1, 0);
	glBegin(GL_LINES);
	float len = 10;
	for(int x = 0; x < TERRAIN_MAX_X; x++)
	{
		for(int z = 0; z < TERRAIN_MAX_Z; z++)
		{
			glVertex3f(x * TILE_SIZE, m_heights[x][z], z * TILE_SIZE);
			glVertex3f(x * TILE_SIZE + m_norms[x][z].x * len, m_heights[x][z] + m_norms[x][z].y * len, z * TILE_SIZE + m_norms[x][z].z * len);
		}
	}
	glEnd();
	glEnable(GL_LIGHTING);
	*/
}

float Terrain::GetHeight(float x, float z) const
{
	float projCameraX = x / TILE_SIZE;
	float projCameraZ = z / TILE_SIZE;

	if(projCameraX < 0 || projCameraX >= TERRAIN_MAX_X || projCameraZ < 0 || projCameraZ >= TERRAIN_MAX_Z)
		return -1;

	int hflCol0 = int(projCameraX);
	int hflRow0 = int(projCameraZ);
	int hflCol1 = hflCol0 + 1;
	int hflRow1 = hflRow0 + 1;

	float h00 = m_heights[hflCol0][hflRow0];
	float h01 = m_heights[hflCol1][hflRow0];
	float h11 = m_heights[hflCol1][hflRow1];
	float h10 = m_heights[hflCol0][hflRow1];

	float tx = projCameraX - float(hflCol0);
	float ty = projCameraZ - float(hflRow0);

	float txty = tx * ty;

	return h00 * (1.0f - ty - tx + txty)
		+ h01 * (tx - txty)
		+ h11 * txty
		+ h10 * (ty - txty);
}

