#pragma once

#include <map>
#include "texture.h"

#include "mathlib/vector.h"


using namespace std;

struct DRAW_TILE
{
	int xmin;
	int xmax;
	int zmin; 
	int zmax;

	DRAW_TILE(int _xmin, int _xmax, int _zmin, int _zmax)
	{
		xmin = _xmin;
		xmax = _xmax;
		zmin = _zmin;
		zmax = _zmax;
	}
};

class Terrain
{

public:
	static const int TERRAIN_MAX_X = 128 + 1;
	static const int TERRAIN_MAX_Z = 128 + 1;
	static const int TILE_SIZE = 10;
	static const int MAX_HEIGHT = 80;

	void Init();
	void Render(const vector3 &mypos);

	float Terrain::GetHeight(float x, float z) const;

private:
	int m_heights[TERRAIN_MAX_X][TERRAIN_MAX_Z];
	float m_lodHeights[TERRAIN_MAX_X][TERRAIN_MAX_Z];
	vector3 m_norms[TERRAIN_MAX_X][TERRAIN_MAX_Z];
	Texture m_terrainTexture;

	typedef multimap<int, DRAW_TILE> DrawSeqMap;
	DrawSeqMap m_drawSeqMap;

	float Terrain::GetTileHeight(int x, int z);

	float Terrain::Displace(float num);
	void Terrain::DivideGrid(int x, int y, int width, int height, float c1, float c2, float c3, float c4);
	
	bool Terrain::CalcLODTerrain(int xmin, int xmax, int zmin, int zmax, int myposx, int myposz, int lev);
	void AddTileLevel(int xmin, int xmax, int zmin, int zmax, int lev);
	void DrawTile(int xmin, int xmax, int zmin, int zmax, int lev);

	inline void Terrain::DrawVert(int x, int z);

};
